/*
 * $Id: picterm.c,v 1.9 2001/02/22 00:35:55 mac Exp $
 * 
 * PIC NIC terminal for linux
 * coded by mac@research.co.jp
 * copyleft (C) 2001 mac@research.co.jp
 *
 */

#include    <stdio.h>
#include    <stdlib.h>
#include    <unistd.h>
#include    <string.h>
#include    <errno.h>
#include    <sys/types.h>
#include    <sys/socket.h>
#include    <sys/param.h>
#include    <netdb.h>
#include    <netinet/in.h>
#include    <arpa/inet.h>
#include    <setjmp.h>
#include    <signal.h>
#include    <wait.h>
#include    <sys/time.h>

#define XTAL (20.0e6)			// 20MHz

void
main(argc, argv)
int                     argc;
char                   *argv[];
{
    FILE                   *fo_pic, *fi_tel;
    char                    buf[BUFSIZ];
    struct sockaddr_in      picsock, telsock;
    struct hostent         *host;
    int                     picfd, telfd, tel;
    int                     i, c;
    int                     foo, on;
    pid_t                   pid;
    unsigned long           myip;
    char                   *pichost, myname[MAXHOSTNAMELEN];
    int                     telport, picport;
    int                     speed;
    char                   *prog;

    prog = argv[0];

    if (argc < 4) {
	fprintf(stderr, "%s picnic picport myport speed\n", prog);
	fprintf(stderr, "\tpicnic: IP or hostname of picnic\n");
	fprintf(stderr, "\tpicport: port # of serial port of picnic\n");
	fprintf(stderr, "\tmyport: port # for listen\n");
	fprintf(stderr, "\tspeed: serial speed (9600 ... 115200)\n");
	exit(0);
    }
    pichost = argv[1];
    picport = atoi(argv[2]);
    telport = atoi(argv[3]);
    speed = atoi(argv[4]);

    if (gethostname(myname, MAXHOSTNAMELEN) < 0) {
	perror("gethostname");
	exit(1);
    }
    if ((host = (struct hostent *) gethostbyname(myname)) == 0) {
	perror("gethostbyname myname");
	exit(1);
    }
    myip = htonl((u_long) * ((int *) host->h_addr));


    picfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (picfd < 0) {
	perror("picterm: socket pic");;
	exit(1);
    }
    fo_pic = fdopen(picfd, "w");

    telfd = socket(AF_INET, SOCK_STREAM, 0);
    telsock.sin_family = AF_INET;
    telsock.sin_port = htons(telport);
    if (telfd < 0) {
	perror("picterm: socket tel");;
	exit(1);
    }
    (void) setsockopt(telfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
    if (bind(telfd, (struct sockaddr *) &telsock, sizeof(telsock)) < 0) {
	perror("bind");
	exit(1);
    }

    if (listen(telfd, 1) < 0) {
	perror("listen");
	exit(1);
    }
    foo = sizeof(telsock);
    tel = accept(telfd, (struct sockaddr *) &telsock, &foo);
    if (tel < 0) {
	perror("accept");
	exit(1);
    }
    fi_tel = fdopen(tel, "r");

    picsock.sin_family = AF_INET;
    picsock.sin_port = htons(picport);

    if (inet_aton(pichost, &picsock.sin_addr)) {
    }
    else if ((host = (struct hostent *) gethostbyname(pichost)) != 0) {
	picsock.sin_addr.s_addr = htonl((u_long) * ((int *) host->h_addr));
    }
    else {
	perror("gethostbyname error");
	exit(errno);
    }

    if (connect(picfd, &picsock, sizeof(picsock)) == -1) {
	perror("connect error");
	exit(errno);
    }

    putc(2, fo_pic);
    fflush(fo_pic);
    i = recv(picfd, buf, BUFSIZ, MSG_OOB);
    putc(0, fo_pic);
    fflush(fo_pic);
    i = recv(picfd, buf, BUFSIZ, MSG_OOB);

    buf[0] = 1;
    buf[1] = (myip >> 24) & 0xff;
    buf[2] = (myip >> 16) & 0xff;
    buf[3] = (myip >> 8) & 0xff;
    buf[4] = myip & 0xff;
    buf[5] = (int) (XTAL / (speed * 16.0) - 0.5);
    buf[6] = 0;

    fwrite(buf, 1, 7, fo_pic);
    fflush(fo_pic);
    if ((pid = fork()) != 0) {
	while ((i = read(tel, buf + 1, BUFSIZ - 1)) > 0) {
	    buf[0] = 3;
	    if (write(picfd, buf, i+1) < 0) {
		perror("tel->pic");
		exit(0);
	    }
	}
	kill(pid, SIGKILL);
	do {
	    pid = waitpid((pid_t) - 1, NULL, WNOHANG);
	    if (pid < 0) {
		exit(0);
	    }
	} while (pid != 0);

	exit(0);
    }
    else {
	while (1) {
	    i = recv(picfd, buf, BUFSIZ, MSG_OOB);
	    if (buf[0] == 4 && buf[1] > 16) {
		if (write(tel, &buf[16], buf[1] - 16) < 0) {
		    exit(0);
		}
	    }
	}
    }
}
